import asyncio

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


async def fetch(url, session, semaphore):
    with (await semaphore):
        try:
            async with session.get(url) as response:
                if response.status in [200, 201]:
                    data = await response.text()
                    return data
                return None
        except:
            return None



import re

from lxml import etree
from pyquery import PyQuery as pq


class Selector:
    def __init__(self, rule, attr=None):
        self.rule = rule
        self.attr = attr

    def __str__(self):
        return '{}({})'.format(self.__class__.__name__, self.rule)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.rule)

    def parse_detail(self, html):
        raise NotImplementedError


class Css(Selector):
    def parse_detail(self, html):
        d = pq(html)
        if self.attr is None:
            try:
                return d(self.rule)[0].text
            except IndexError:
                return None
        return d(self.rule)[0].attr(self.attr, None)


class Xpath(Selector):
    def parse_detail(self, html):
        d = etree.HTML(html)
        try:
            if self.attr is None:
                return d.xpath(self.rule)[0].text
            return d.xpath(self.rule)[0].get(self.attr, None)
        except IndexError:
            return None


class Regex(Selector):
    def parse_detail(self, html):
        try:
            return re.findall(self.rule, html)[0]
        except IndexError:
            return None



import asyncio
import re
from pybloomfilter import BloomFilter

import aiohttp

from gain.request import fetch
from .log import logger


class Parser:
    def __init__(self, rule, item=None):
        self.rule = rule
        self.item = item
        self.parsing_urls = []
        self.pre_parse_urls = []
        self.filter_urls = BloomFilter(10000000, 0.01)
        self.done_urls = []

    def add(self, urls):
        url = '{}'.format(urls)
        if url.encode('utf-8') not in self.filter_urls:
            self.filter_urls.add(url.encode('utf-8'))
            self.pre_parse_urls.append(url)

    def parse_urls(self, html):
        urls = re.findall(self.rule, html)
        for url in urls:
            self.add(url)

    async def parse_item(self, html):
        item = self.item(html)
        await item.save()
        self.item._item_count += 1
        return item

    async def execute_url(self, spider, session, semaphore, url):
        html = await fetch(url, session, semaphore)

        if html is None:
            spider.error_urls.append(url)
            self.pre_parse_urls.append(url)
            return
        if url in spider.error_urls:
            spider.error_urls.remove(url)
        spider.urls_count += 1
        self.parsing_urls.remove(url)
        self.done_urls.append(url)
        if self.item is not None:
            await self.parse_item(html)
            logger.info('Parsed({}/{}): {}'.format(len(self.done_urls), len(self.filter_urls), url))
        else:
            spider.parse(html)
            logger.info('Followed({}/{}): {}'.format(len(self.done_urls), len(self.filter_urls), url))

    async def task(self, spider, semaphore):
        with aiohttp.ClientSession() as session:
            while spider.is_running():
                if len(self.pre_parse_urls) == 0:
                    await asyncio.sleep(0.5)
                    continue
                url = self.pre_parse_urls.pop()
                self.parsing_urls.append(url)
                asyncio.ensure_future(self.execute_url(spider, session, semaphore, url))



import logging

logging.basicConfig(level=logging.DEBUG,
                    format='[%(asctime)s] %(message)s',
                    datefmt='%Y:%m:%d %H:%M:%S')
logger = logging.getLogger('running')



from .log import logger
from .selector import Selector


class ItemType(type):
    def __new__(mcs, name, bases, namespace):
        selectors = {}
        namespace['_item_name'] = name
        namespace['_item_count'] = 0
        for name, value in namespace.items():
            if isinstance(value, Selector):
                selectors[name] = value
        namespace['selectors'] = selectors
        for name, value in selectors.items():
            del namespace[name]
        return type.__new__(mcs, name, bases, namespace)


class Item(metaclass=ItemType):
    def __init__(self, html):
        self.results = {}
        for name, selector in self.selectors.items():
            value = selector.parse_detail(html)
            if value is None:
                logger.error('Selector "{}" for {} was wrong, please check again'.format(selector.rule, name))
            else:
                self.results[name] = value

    def __getattr__(self, item):
        if item not in self.results:
            raise AttributeError
        return self.results[item]

    async def save(self):
        raise NotImplementedError



from setuptools import find_packages, setup

setup(
    name="gain",
    version="0.1.2",
    description="Web crawling framework for everyone.",
    author="Gaojiuli",
    author_email="gaojiuli@gmail.com",
    url='https://github.com/gaojiuli/gain',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'aiohttp',
        'pybloomfiltermmap',
        'pyquery',
    ],
    license='GNU GPL 3',
    packages=find_packages(),
    py_modules=['gain'],
    include_package_data=True,
    zip_safe=False
)



import asyncio
from datetime import datetime

import aiohttp

from gain.request import fetch
from .log import logger

try:
    import uvloop

    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    pass


class Spider:
    start_url = ''
    parsers = []
    error_urls = []
    urls_count = 0
    concurrency = 5

    @classmethod
    def is_running(cls):
        is_running = False
        for parser in cls.parsers:
            if len(parser.pre_parse_urls) > 0 or len(parser.parsing_urls) > 0:
                is_running = True
        return is_running

    @classmethod
    def parse(cls, html):
        for parser in cls.parsers:
            parser.parse_urls(html)

    @classmethod
    def run(cls):
        logger.info('Spider started!')
        start_time = datetime.now()
        semaphore = asyncio.Semaphore(cls.concurrency)
        loop = asyncio.get_event_loop()

        tasks = asyncio.wait([parser.task(cls, semaphore) for parser in cls.parsers])

        try:
            loop.run_until_complete(cls.init_parse(semaphore))
            loop.run_until_complete(tasks)
        except KeyboardInterrupt:
            for task in asyncio.Task.all_tasks():
                task.cancel()
            loop.run_forever()
        finally:
            end_time = datetime.now()
            for parser in cls.parsers:
                if parser.item is not None:
                    logger.info('Item "{}": {}'.format(parser.item._item_name, parser.item._item_count))
            logger.info('Requests count: {}'.format(cls.urls_count))
            logger.info('Error count: {}'.format(len(cls.error_urls)))
            logger.info('Time usage: {}'.format(end_time - start_time))
            logger.info('Spider finished!')
            loop.close()

    @classmethod
    async def init_parse(cls, semaphore):
        with aiohttp.ClientSession() as session:
            html = await fetch(cls.start_url, session, semaphore)
            cls.parse(html)
